# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    List,
    Optional,
)
import dearpygui.dearpygui as dpg
from numpy import (
    array,
    ceil,
    floor,
    log10 as log,
    ndarray,
)
import deareis.themes as themes
from deareis.gui.plots.base import Plot


class Residuals(Plot):
    def __init__(self, width: int = -1, height: int = -1, *args, **kwargs):
        assert type(width) is int, width
        assert type(height) is int, height
        super().__init__()
        with dpg.plot(
            anti_aliased=True,
            crosshairs=True,
            width=width,
            height=height,
            tag=self._plot,
        ):
            dpg.add_plot_legend(
                horizontal=True,
                location=dpg.mvPlot_Location_North,
                outside=kwargs.get("legend_outside", True),
            )
            self._x_axis: int = dpg.add_plot_axis(
                dpg.mvXAxis,
                label="f (Hz)",
                log_scale=True,
                no_gridlines=True,
            )
            self._y_axis_1: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label="Z' error (%)",
            )
            self._y_axis_2: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label='Z" error (%)',
            )
        dpg.bind_item_theme(self._plot, themes.plot)
        dpg.bind_item_handler_registry(self._plot, self._item_handler)

    @classmethod
    def duplicate(Class, original, *args, **kwargs) -> Plot:
        copy: Plot = Class(*args, **kwargs)
        for kwargs in original.get_series():
            copy.plot(**kwargs)
        return copy

    def is_blank(self) -> bool:
        return (
            len(dpg.get_item_children(self._y_axis_1, slot=1)) == 0
            and len(dpg.get_item_children(self._y_axis_2, slot=1)) == 0
        )

    def clear(self, *args, **kwargs):
        delete: bool = kwargs.get("delete", True)
        if delete:
            dpg.delete_item(self._y_axis_1, children_only=True)
            dpg.delete_item(self._y_axis_2, children_only=True)
            self._series.clear()
        else:
            i: int
            series_1: int
            series_2: int
            for i, (series_1, series_2) in enumerate(
                zip(
                    dpg.get_item_children(self._y_axis_1, slot=1),
                    dpg.get_item_children(self._y_axis_2, slot=1),
                )
            ):
                if i < len(self._series):
                    self._series[i]["frequency"] = array([])
                    self._series[i]["real"] = array([])
                    self._series[i]["imaginary"] = array([])
                dpg.set_value(series_1, [[], []])
                dpg.set_value(series_2, [[], []])

    def update(self, index: int, *args, **kwargs):
        assert type(index) is int and index >= 0, index
        assert len(self._series) > index, (
            index,
            len(self._series),
        )
        assert len(args) == 0, args
        freq: ndarray = kwargs["frequency"]
        real: ndarray = kwargs["real"]
        imag: ndarray = kwargs["imaginary"]
        assert type(freq) is ndarray, freq
        assert type(real) is ndarray, real
        assert type(imag) is ndarray, imag
        i: int
        series_1: int
        series_2: int
        for i, (series_1, series_2) in enumerate(
            zip(
                dpg.get_item_children(self._y_axis_1, slot=1),
                dpg.get_item_children(self._y_axis_2, slot=1),
            )
        ):
            if not (i == index or i == index + 1):
                continue
            if i == index:
                self._series[index].update(kwargs)
            dpg.set_value(series_1, [list(freq), list(real)])
            dpg.set_value(series_2, [list(freq), list(imag)])

    def plot(self, *args, **kwargs):
        assert len(args) == 0, args
        freq: ndarray = kwargs["frequency"]
        real: ndarray = kwargs["real"]
        imag: ndarray = kwargs["imaginary"]
        assert type(freq) is ndarray, freq
        assert type(real) is ndarray, real
        assert type(imag) is ndarray, imag
        self._series.append(kwargs)
        x: list = list(freq)
        y: list = list(real)
        dpg.bind_item_theme(
            dpg.add_scatter_series(
                x=x,
                y=y,
                label="Z'",
                user_data=(
                    freq,
                    real,
                    imag,
                ),
                parent=self._y_axis_1,
            ),
            themes.residuals.real,
        )
        dpg.bind_item_theme(
            dpg.add_line_series(
                x=x,
                y=y,
                user_data=(
                    freq,
                    real,
                    imag,
                ),
                parent=self._y_axis_1,
            ),
            themes.residuals.real,
        )
        y = list(imag)
        dpg.bind_item_theme(
            dpg.add_scatter_series(
                x=x,
                y=y,
                label='Z"',
                parent=self._y_axis_2,
            ),
            themes.residuals.imaginary,
        )
        dpg.bind_item_theme(
            dpg.add_line_series(
                x=x,
                y=y,
                parent=self._y_axis_2,
            ),
            themes.residuals.imaginary,
        )

    def adjust_limits(self):
        if not self.is_visible():
            self.queue_limits_adjustment()
            return
        elif self.are_limits_adjusted():
            return
        else:
            self.limits_adjusted()
        error_lim: float = 0.0
        freq: Optional[ndarray] = None
        for kwargs in self._series:
            freq = kwargs["frequency"]
            if not freq.size > 0:
                continue
            real: ndarray = kwargs["real"]
            imag: ndarray = kwargs["imaginary"]
            if max(real) > error_lim:
                error_lim = max(real)
            if abs(min(real)) > error_lim:
                error_lim = abs(min(real))
            if max(imag) > error_lim:
                error_lim = max(imag)
            if abs(min(imag)) > error_lim:
                error_lim = abs(min(imag))
        if error_lim <= 0.5:
            if 0.5 - error_lim > 0.1:
                error_lim = 0.5
            else:
                error_lim = 1.0
        elif error_lim <= 5:
            if ceil(error_lim) - error_lim > 0.1:
                error_lim = ceil(error_lim)
            else:
                error_lim = ceil(error_lim) + 1.0
        else:
            n: int = 5
            error_lim = ceil(error_lim / n) * n + n
        dpg.split_frame()
        min_x: float = min(freq) if freq.size > 0 else 0.5
        max_x: float = max(freq) if freq.size > 0 else 1.0
        if min_x == 0.0:
            min_x = 1e-1
        dx: float = 0.1
        dpg.set_axis_limits(
            self._x_axis,
            ymin=10 ** (floor(log(min_x) / dx) * dx - dx)
            if (freq is not None and freq.size > 0)
            else 0,
            ymax=10 ** (ceil(log(max_x) / dx) * dx + dx)
            if (freq is not None and freq.size > 0)
            else 1,
        )
        dpg.set_axis_limits(self._y_axis_1, ymin=-error_lim, ymax=error_lim)
        dpg.set_axis_limits(self._y_axis_2, ymin=-error_lim, ymax=error_lim)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._y_axis_1)
        dpg.set_axis_limits_auto(self._y_axis_2)

    def copy_limits(self, other: Plot):
        src: int
        dst: int
        for src, dst in zip(
            [
                other._x_axis,
                other._y_axis_1,
                other._y_axis_2,
            ],
            [
                self._x_axis,
                self._y_axis_1,
                self._y_axis_2,
            ],
        ):
            limits: List[float] = dpg.get_axis_limits(src)
            dpg.set_axis_limits(dst, *limits)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis_1)
        dpg.set_axis_limits_auto(self._y_axis_2)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from time import time as _time
from typing import Optional
from uuid import uuid4 as _uuid4
from numpy import (
    integer as _integer,
    issubdtype as _issubdtype,
)
import pyimpspec as _pyimpspec
from pyimpspec import (
    Circuit,
    FittedParameter,
    FittingError,
)
from deareis.data import (
    DataSet,
    FitResult,
    FitSettings,
)
from deareis.enums import (
    CNLSMethod,
    Weight,
    value_to_cnls_method as _value_to_cnls_method,
    value_to_weight as _value_to_weight,
    cnls_method_to_value as _cnls_method_to_value,
    weight_to_value as _weight_to_value,
)


def fit_circuit(
    data: DataSet,
    settings: FitSettings,
    num_procs: int = -1,
) -> FitResult:
    """
    Wrapper for the `pyimpspec.fit_circuit` function.

    Fit a circuit to a data set.

    Parameters
    ----------
    data: DataSet
        The data set that the circuit will be fitted to.

    settings: FitSettings
        The settings that determine the circuit and how the fit is performed.

    num_procs: int = -1
        The maximum number of parallel processes to use when method is `CNLSMethod.AUTO` and/or weight is `Weight.AUTO`.

    Returns
    -------
    FitResult
    """
    assert isinstance(data, _pyimpspec.DataSet), data
    assert type(settings) is FitSettings, settings
    assert _issubdtype(type(num_procs), _integer), num_procs
    circuit: Circuit = _pyimpspec.parse_cdc(settings.cdc)
    result: _pyimpspec.FitResult = _pyimpspec.fit_circuit(
        circuit=circuit,
        data=data,
        method=_cnls_method_to_value.get(settings.method, "auto"),
        weight=_weight_to_value.get(settings.weight, "auto"),
        max_nfev=settings.max_nfev,
        num_procs=num_procs,
    )
    method: Optional[CNLSMethod] = _value_to_cnls_method.get(result.method)
    weight: Optional[Weight] = _value_to_weight.get(result.weight)
    assert method is not None
    assert weight is not None
    return FitResult(
        _uuid4().hex,
        _time(),
        result.circuit,
        result.parameters,
        result.frequency,
        result.impedance,
        result.real_residual,
        result.imaginary_residual,
        data.get_mask(),
        result.minimizer_result.chisqr,
        result.minimizer_result.redchi,
        result.minimizer_result.aic,
        result.minimizer_result.bic,
        result.minimizer_result.ndata,
        result.minimizer_result.nfree,
        result.minimizer_result.nfev,
        method,
        weight,
        settings,
    )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Dict,
    List,
)
import dearpygui.dearpygui as dpg
from deareis.enums import (
    Action,
    DRTMethod,
    DRTMode,
    CNLSMethod,
    TestMode,
    PlotLegendLocation,
    PlotPreviewLimit,
    PlotUnits,
    RBFShape,
    RBFType,
    Test,
    Weight,
)
from deareis.keybindings import Keybinding
from deareis.data.plotting import PlotExportSettings
from deareis.data import (
    DRTSettings,
    FitSettings,
    SimulationSettings,
    TestSettings,
)

DEFAULT_KEYBINDINGS: List[Keybinding] = [
    Keybinding(
        dpg.mvKey_N,
        False,
        True,
        False,
        Action.NEW_PROJECT,
    ),
    Keybinding(
        dpg.mvKey_O,
        False,
        True,
        False,
        Action.LOAD_PROJECT,
    ),
    Keybinding(
        dpg.mvKey_Q,
        False,
        True,
        False,
        Action.EXIT,
    ),
    Keybinding(
        dpg.mvKey_Next,
        False,
        False,
        True,
        Action.NEXT_PROGRAM_TAB,
    ),
    Keybinding(
        dpg.mvKey_Prior,
        False,
        False,
        True,
        Action.PREVIOUS_PROGRAM_TAB,
    ),
    Keybinding(
        dpg.mvKey_F1,
        False,
        False,
        False,
        Action.SHOW_SETTINGS_APPEARANCE,
    ),
    Keybinding(
        dpg.mvKey_F2,
        False,
        False,
        False,
        Action.SHOW_SETTINGS_DEFAULTS,
    ),
    Keybinding(
        dpg.mvKey_F3,
        False,
        False,
        False,
        Action.SHOW_SETTINGS_KEYBINDINGS,
    ),
    Keybinding(
        dpg.mvKey_F11,
        False,
        False,
        False,
        Action.SHOW_HELP_LICENSES,
    ),
    Keybinding(
        dpg.mvKey_F12,
        False,
        False,
        False,
        Action.SHOW_HELP_ABOUT,
    ),
    Keybinding(
        dpg.mvKey_P,
        False,
        True,
        False,
        Action.SHOW_COMMAND_PALETTE,
    ),
    Keybinding(
        dpg.mvKey_S,
        False,
        True,
        False,
        Action.SAVE_PROJECT,
    ),
    Keybinding(
        dpg.mvKey_S,
        False,
        True,
        True,
        Action.SAVE_PROJECT_AS,
    ),
    Keybinding(
        dpg.mvKey_W,
        False,
        True,
        False,
        Action.CLOSE_PROJECT,
    ),
    Keybinding(
        dpg.mvKey_Z,
        False,
        True,
        False,
        Action.UNDO,
    ),
    Keybinding(
        dpg.mvKey_Y if dpg.get_platform() == dpg.mvPlatform_Windows else dpg.mvKey_Z,
        False,
        True,
        False if dpg.get_platform() == dpg.mvPlatform_Windows else True,
        Action.REDO,
    ),
    Keybinding(
        dpg.mvKey_Next,
        False,
        True,
        False,
        Action.NEXT_PROJECT_TAB,
    ),
    Keybinding(
        dpg.mvKey_Prior,
        False,
        True,
        False,
        Action.PREVIOUS_PROJECT_TAB,
    ),
    Keybinding(
        dpg.mvKey_D,
        True,
        True,
        False,
        Action.SELECT_DATA_SETS_TAB,
    ),
    Keybinding(
        dpg.mvKey_F,
        True,
        True,
        False,
        Action.SELECT_FITTING_TAB,
    ),
    Keybinding(
        dpg.mvKey_K,
        True,
        True,
        False,
        Action.SELECT_KRAMERS_KRONIG_TAB,
    ),
    Keybinding(
        dpg.mvKey_O,
        True,
        True,
        False,
        Action.SELECT_OVERVIEW_TAB,
    ),
    Keybinding(
        dpg.mvKey_P,
        True,
        True,
        False,
        Action.SELECT_PLOTTING_TAB,
    ),
    Keybinding(
        dpg.mvKey_S,
        True,
        True,
        False,
        Action.SELECT_SIMULATION_TAB,
    ),
    Keybinding(
        dpg.mvKey_Return,
        False if dpg.get_platform() == dpg.mvPlatform_Windows else True,
        True if dpg.get_platform() == dpg.mvPlatform_Windows else False,
        False,
        Action.PERFORM_ACTION,
    ),
    Keybinding(
        dpg.mvKey_Delete,
        True,
        False,
        False,
        Action.DELETE_RESULT,
    ),
    Keybinding(
        dpg.mvKey_Next,
        False,
        False,
        False,
        Action.NEXT_PRIMARY_RESULT,
    ),
    Keybinding(
        dpg.mvKey_Prior,
        False,
        False,
        False,
        Action.PREVIOUS_PRIMARY_RESULT,
    ),
    Keybinding(
        dpg.mvKey_Next,
        True,
        False,
        False,
        Action.NEXT_SECONDARY_RESULT,
    ),
    Keybinding(
        dpg.mvKey_Prior,
        True,
        False,
        False,
        Action.PREVIOUS_SECONDARY_RESULT,
    ),
    Keybinding(
        dpg.mvKey_A,
        True,
        False,
        False,
        Action.APPLY_SETTINGS,
    ),
    Keybinding(
        dpg.mvKey_M,
        True,
        False,
        False,
        Action.APPLY_MASK,
    ),
    Keybinding(
        dpg.mvKey_N,
        True,
        False,
        False,
        Action.SHOW_ENLARGED_NYQUIST,
    ),
    Keybinding(
        dpg.mvKey_D,
        True,
        False,
        False,
        Action.SHOW_ENLARGED_DRT,
    ),
    Keybinding(
        dpg.mvKey_I,
        True,
        False,
        False,
        Action.SHOW_ENLARGED_IMPEDANCE,
    ),
    Keybinding(
        dpg.mvKey_B,
        True,
        False,
        False,
        Action.SHOW_ENLARGED_BODE,
    ),
    Keybinding(
        dpg.mvKey_R,
        True,
        False,
        False,
        Action.SHOW_ENLARGED_RESIDUALS,
    ),
    Keybinding(
        dpg.mvKey_E,
        True,
        False,
        False,
        Action.SHOW_CIRCUIT_EDITOR,
    ),
    Keybinding(
        dpg.mvKey_D,
        True,
        False,
        True,
        Action.COPY_DRT_DATA,
    ),
    Keybinding(
        dpg.mvKey_I,
        True,
        False,
        True,
        Action.COPY_IMPEDANCE_DATA,
    ),
    Keybinding(
        dpg.mvKey_N,
        True,
        False,
        True,
        Action.COPY_NYQUIST_DATA,
    ),
    Keybinding(
        dpg.mvKey_B,
        True,
        False,
        True,
        Action.COPY_BODE_DATA,
    ),
    Keybinding(
        dpg.mvKey_R,
        True,
        False,
        True,
        Action.COPY_RESIDUALS_DATA,
    ),
    Keybinding(
        dpg.mvKey_C,
        True,
        False,
        False,
        Action.COPY_OUTPUT,
    ),
    Keybinding(
        dpg.mvKey_A,
        True,
        False,
        False,
        Action.AVERAGE_DATA_SETS,
    ),
    Keybinding(
        dpg.mvKey_T,
        True,
        False,
        False,
        Action.TOGGLE_DATA_POINTS,
    ),
    Keybinding(
        dpg.mvKey_C,
        True,
        False,
        False,
        Action.COPY_DATA_SET_MASK,
    ),
    Keybinding(
        dpg.mvKey_S,
        True,
        False,
        False,
        Action.SUBTRACT_IMPEDANCE,
    ),
    Keybinding(
        dpg.mvKey_A,
        True,
        False,
        False,
        Action.SELECT_ALL_PLOT_SERIES,
    ),
    Keybinding(
        dpg.mvKey_A,
        True,
        False,
        True,
        Action.UNSELECT_ALL_PLOT_SERIES,
    ),
    Keybinding(
        dpg.mvKey_C,
        True,
        False,
        False,
        Action.COPY_PLOT_APPEARANCE,
    ),
    Keybinding(
        dpg.mvKey_C,
        True,
        False,
        True,
        Action.COPY_PLOT_DATA,
    ),
    Keybinding(
        dpg.mvKey_E,
        True,
        False,
        False,
        Action.EXPAND_COLLAPSE_SIDEBAR,
    ),
    Keybinding(
        dpg.mvKey_P,
        True,
        False,
        False,
        Action.EXPORT_PLOT,
    ),
    Keybinding(
        dpg.mvKey_L,
        True,
        False,
        False,
        Action.LOAD_SIMULATION_AS_DATA_SET,
    ),
]

# TODO: Replace string keys with int keys (e.g. themes.nyquist.data) and use strings only in the
# config file
DEFAULT_MARKERS: Dict[str, int] = {
    "bode_magnitude_data": dpg.mvPlotMarker_Circle,
    "bode_magnitude_simulation": dpg.mvPlotMarker_Cross,
    "bode_phase_data": dpg.mvPlotMarker_Square,
    "bode_phase_simulation": dpg.mvPlotMarker_Plus,
    "impedance_imaginary_data": dpg.mvPlotMarker_Square,
    "impedance_imaginary_simulation": dpg.mvPlotMarker_Plus,
    "impedance_real_data": dpg.mvPlotMarker_Circle,
    "impedance_real_simulation": dpg.mvPlotMarker_Cross,
    "mu_Xps_Xps": dpg.mvPlotMarker_Square,
    "mu_Xps_mu": dpg.mvPlotMarker_Circle,
    "nyquist_data": dpg.mvPlotMarker_Circle,
    "nyquist_simulation": dpg.mvPlotMarker_Cross,
    "residuals_imaginary": dpg.mvPlotMarker_Square,
    "residuals_real": dpg.mvPlotMarker_Circle,
}

DEFAULT_COLORS: Dict[str, List[float]] = {
    "residuals_real": [
        238.0,
        51.0,
        119.0,
        190.0,
    ],
    "residuals_imaginary": [
        0.0,
        153.0,
        136.0,
        190.0,
    ],
    "nyquist_data": [
        51.0,
        187.0,
        238.0,
        190.0,
    ],
    "nyquist_simulation": [
        238.0,
        51.0,
        119.0,
        190.0,
    ],
    "bode_magnitude_data": [
        51.0,
        187.0,
        238.0,
        190.0,
    ],
    "bode_magnitude_simulation": [
        238.0,
        51.0,
        119.0,
        190.0,
    ],
    "bode_phase_data": [
        238.0,
        119.0,
        51.0,
        190.0,
    ],
    "bode_phase_simulation": [
        0.0,
        153.0,
        136.0,
        190.0,
    ],
    "mu_Xps_mu_criterion": [
        255.0,
        255.0,
        255.0,
        128.0,
    ],
    "mu_Xps_mu": [
        238.0,
        51.0,
        119.0,
        190.0,
    ],
    "mu_Xps_mu_highlight": [
        51.0,
        187.0,
        238.0,
        190.0,
    ],
    "mu_Xps_Xps": [
        0.0,
        153.0,
        136.0,
        190.0,
    ],
    "mu_Xps_Xps_highlight": [
        238.0,
        119.0,
        51.0,
        190.0,
    ],
    "impedance_real_data": [
        51.0,
        187.0,
        238.0,
        190.0,
    ],
    "impedance_real_simulation": [
        238.0,
        51.0,
        119.0,
        190.0,
    ],
    "impedance_imaginary_data": [
        238.0,
        119.0,
        51.0,
        190.0,
    ],
    "impedance_imaginary_simulation": [
        0.0,
        153.0,
        136.0,
        190.0,
    ],
    "drt_real_gamma": [
        51.0,
        187.0,
        238.0,
        190.0,
    ],
    "drt_mean_gamma": [
        238.0,
        119.0,
        51.0,
        190.0,
    ],
    "drt_credible_intervals": [
        238.0,
        119.0,
        51.0,
        48.0,
    ],
    "drt_imaginary_gamma": [
        238.0,
        119.0,
        51.0,
        190.0,
    ],
}

DEFAULT_TEST_SETTINGS: TestSettings = TestSettings(
    test=Test.COMPLEX,
    mode=TestMode.EXPLORATORY,
    num_RC=999,
    mu_criterion=0.85,
    add_capacitance=True,
    add_inductance=True,
    method=CNLSMethod.LEASTSQ,
    max_nfev=1000,
)

DEFAULT_FIT_SETTINGS: FitSettings = FitSettings(
    cdc="R([RW]C)",
    method=CNLSMethod.AUTO,
    weight=Weight.AUTO,
    max_nfev=1000,
)

DEFAULT_SIMULATION_SETTINGS: SimulationSettings = SimulationSettings(
    cdc="R([RW]C)",
    min_frequency=1.0e-2,
    max_frequency=1.0e5,
    num_per_decade=1,
)

DEFAULT_DRT_SETTINGS: DRTSettings = DRTSettings(
    method=DRTMethod.TR_RBF,
    mode=DRTMode.COMPLEX,
    lambda_value=1e-3,
    rbf_type=RBFType.GAUSSIAN,
    derivative_order=1,
    rbf_shape=RBFShape.FWHM,
    shape_coeff=0.5,
    inductance=False,
    credible_intervals=False,
    num_samples=2000,
    num_attempts=10,
    maximum_symmetry=0.5,
    circuit=None,
    W=0.15,
    num_per_decade=100,
)

DEFAULT_PLOT_EXPORT_SETTINGS: PlotExportSettings = PlotExportSettings(
    units=PlotUnits.INCHES,
    width=10.0,
    height=6.0,
    dpi=100,
    preview_limit=PlotPreviewLimit.PX1024,
    show_title=True,
    show_legend=True,
    legend_location=PlotLegendLocation.AUTO,
    show_grid=False,
    has_tight_layout=True,
    num_per_decade=100,
    extension=".png",
    clear_registry=True,
    disable_preview=False,
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from traceback import format_exc
from typing import List, Optional, Set
import dearpygui.dearpygui as dpg
import deareis.signals as signals
from deareis.signals import Signal
from .keybinding import Keybinding, dpg_to_string
from deareis.enums import Action, Context, action_contexts
from deareis.data import DataSet, TestResult, FitResult, SimulationResult


def is_shift_down() -> bool:
    return dpg.is_key_down(dpg.mvKey_Shift) or dpg.is_key_down(dpg.mvKey_RShift)


def is_control_down() -> bool:
    return (
        dpg.is_key_down(dpg.mvKey_Control)
        or dpg.is_key_down(dpg.mvKey_LControl)
        or dpg.is_key_down(dpg.mvKey_RControl)
    )


def is_alt_down() -> bool:
    return dpg.is_key_down(dpg.mvKey_Alt)


class KeybindingHandler:
    def __init__(self, keybindings: List[Keybinding], state):
        self.block_events: bool = False
        self.state = state
        self.keybindings: List[Keybinding] = keybindings
        if not keybindings:
            return
        self.key_handler: int = -1
        self.register(self.keybindings)

    def register(self, keybindings: List[Keybinding]):
        self.keybindings = keybindings
        registered_keys: Set[int] = set()
        if self.key_handler > 0 and dpg.does_item_exist(self.key_handler):
            dpg.delete_item(self.key_handler)
        self.key_handler = dpg.generate_uuid()
        with dpg.handler_registry(tag=self.key_handler):
            kb: Keybinding
            for kb in keybindings:
                if kb.key in registered_keys:
                    continue
                registered_keys.add(kb.key)
                dpg.add_key_release_handler(
                    key=kb.key,
                    callback=lambda s, a, u: self.process(a),
                )

    def block(self):
        self.block_events = True

    def unblock(self):
        self.block_events = False

    def process(self, key: int):
        if self.block_events:
            # TODO: Remove this workaround when no longer required by the file dialog window?
            window: Optional[int] = self.state.get_active_modal_window()
            if (
                window is None
                or not dpg.does_item_exist(window)
                or not dpg.is_item_shown(window)
            ):
                signals.emit(Signal.UNBLOCK_KEYBINDINGS)
            else:
                return
        filtered_keybindings: List[Keybinding] = list(
            filter(lambda _: key in _, self.keybindings)
        )
        if not filtered_keybindings:
            return
        if is_alt_down():
            filtered_keybindings = list(
                filter(lambda _: _.mod_alt is True, filtered_keybindings)
            )
        else:
            filtered_keybindings = list(
                filter(lambda _: _.mod_alt is False, filtered_keybindings)
            )
        if not filtered_keybindings:
            return
        if is_control_down():
            filtered_keybindings = list(
                filter(lambda _: _.mod_ctrl is True, filtered_keybindings)
            )
        else:
            filtered_keybindings = list(
                filter(lambda _: _.mod_ctrl is False, filtered_keybindings)
            )
        if not filtered_keybindings:
            return
        if is_shift_down():
            filtered_keybindings = list(
                filter(lambda _: _.mod_shift is True, filtered_keybindings)
            )
        else:
            filtered_keybindings = list(
                filter(lambda _: _.mod_shift is False, filtered_keybindings)
            )
        if not filtered_keybindings:
            return
        project = self.state.get_active_project()  # Optional[Project]
        project_tab = self.state.get_active_project_tab()  # Optional[ProjectTab]
        context: Context
        if project is not None and project_tab is not None:
            context = project_tab.get_active_context()
            filtered_keybindings = list(
                filter(
                    lambda _: context in action_contexts[_.action]
                    or Context.PROJECT in action_contexts[_.action]
                    or Context.PROGRAM in action_contexts[_.action],
                    filtered_keybindings,
                )
            )
        else:
            filtered_keybindings = list(
                filter(
                    lambda _: Context.PROGRAM in action_contexts[_.action],
                    filtered_keybindings,
                )
            )
        if not filtered_keybindings:
            return
        context = (
            project_tab.get_active_context()
            if project_tab is not None
            else Context.PROGRAM
        )
        # Check for active input fields
        if project_tab is not None and project_tab.has_active_input(context):
            return
        try:
            self.validate_keybindings(filtered_keybindings)
            self.perform_action(
                filtered_keybindings[0].action,
                context,
                project,
                project_tab,
            )
        except Exception:
            signals.emit(Signal.SHOW_ERROR_MESSAGE, traceback=format_exc())

    def validate_keybindings(self, keybindings: List[Keybinding]):
        assert len(set(list(map(str, keybindings)))) == 1, (
            "The same keybinding has been applied to multiple actions:\n- "
            + "\n- ".join(list(map(str, keybindings)))
        )

    def perform_action(
        self,
        action: Action,
        context: Context,
        project,  # Optional["Project"]
        project_tab,  # Optional["ProjectTab"]
    ):
        assert type(action) is Action, action
        assert type(context) is Context, context
        # Program-level
        if action == Action.NEW_PROJECT:
            signals.emit(Signal.NEW_PROJECT)
        elif action == Action.LOAD_PROJECT:
            signals.emit(Signal.SELECT_PROJECT_FILES)
        elif action == Action.EXIT:
            dpg.stop_dearpygui()
        elif action == Action.NEXT_PROGRAM_TAB:
            self.state.program_window.select_next_tab()
        elif action == Action.PREVIOUS_PROGRAM_TAB:
            self.state.program_window.select_previous_tab()
        elif action == Action.SELECT_HOME_TAB:
            self.state.program_window.select_home_tab()
        elif action == Action.SHOW_HELP_ABOUT:
            signals.emit(Signal.SHOW_HELP_ABOUT)
        elif action == Action.SHOW_HELP_LICENSES:
            signals.emit(Signal.SHOW_HELP_LICENSES)
        elif action == Action.SHOW_SETTINGS_APPEARANCE:
            signals.emit(Signal.SHOW_SETTINGS_APPEARANCE)
        elif action == Action.SHOW_SETTINGS_DEFAULTS:
            signals.emit(Signal.SHOW_SETTINGS_DEFAULTS)
        elif action == Action.SHOW_SETTINGS_KEYBINDINGS:
            signals.emit(Signal.SHOW_SETTINGS_KEYBINDINGS)
        elif action == Action.SHOW_COMMAND_PALETTE:
            signals.emit(Signal.SHOW_COMMAND_PALETTE)
        elif action == Action.SHOW_CHANGELOG:
            signals.emit(Signal.SHOW_CHANGELOG)
        elif action == Action.CHECK_UPDATES:
            signals.emit(Signal.CHECK_UPDATES)
        elif project is not None:
            assert project is not None
            assert project_tab is not None
            # Project-level
            test: Optional[TestResult]
            fit: Optional[FitResult]
            if action == Action.SAVE_PROJECT:
                signals.emit(Signal.SAVE_PROJECT)
            elif action == Action.SAVE_PROJECT_AS:
                signals.emit(Signal.SAVE_PROJECT_AS)
            elif action == Action.CLOSE_PROJECT:
                signals.emit(Signal.CLOSE_PROJECT)
            elif action == Action.UNDO:
                if not (
                    context == Context.OVERVIEW_TAB
                    and dpg.is_item_focused(project_tab.overview_tab.notes_input)
                ):
                    signals.emit(Signal.UNDO_PROJECT_ACTION)
            elif action == Action.REDO:
                signals.emit(Signal.REDO_PROJECT_ACTION)
            elif action == Action.NEXT_PROJECT_TAB:
                project_tab.select_next_tab()
            elif action == Action.PREVIOUS_PROJECT_TAB:
                project_tab.select_previous_tab()
            elif action == Action.SELECT_OVERVIEW_TAB:
                project_tab.select_overview_tab()
            elif action == Action.SELECT_DATA_SETS_TAB:
                project_tab.select_data_sets_tab()
            elif action == Action.SELECT_KRAMERS_KRONIG_TAB:
                project_tab.select_kramers_kronig_tab()
            elif action == Action.SELECT_DRT_TAB:
                project_tab.select_drt_tab()
            elif action == Action.SELECT_FITTING_TAB:
                project_tab.select_fitting_tab()
            elif action == Action.SELECT_SIMULATION_TAB:
                project_tab.select_simulation_tab()
            elif action == Action.SELECT_PLOTTING_TAB:
                project_tab.select_plotting_tab()
            # Project-level: multiple tabs
            elif action == Action.PERFORM_ACTION:
                if context == Context.DATA_SETS_TAB:
                    signals.emit(Signal.SELECT_DATA_SET_FILES)
                elif context == Context.KRAMERS_KRONIG_TAB:
                    signals.emit(
                        Signal.PERFORM_TEST,
                        data=project_tab.get_active_data_set(),
                        settings=project_tab.get_test_settings(),
                    ),
                elif context == Context.DRT_TAB:
                    signals.emit(
                        Signal.PERFORM_DRT,
                        data=project_tab.get_active_data_set(),
                        settings=project_tab.get_drt_settings(),
                    )
                elif context == Context.FITTING_TAB:
                    signals.emit(
                        Signal.PERFORM_FIT,
                        data=project_tab.get_active_data_set(),
                        settings=project_tab.get_fit_settings(),
                    ),
                elif context == Context.SIMULATION_TAB:
                    signals.emit(
                        Signal.PERFORM_SIMULATION,
                        data=project_tab.get_active_data_set(),
                        settings=project_tab.get_simulation_settings(),
                    ),
                elif context == Context.PLOTTING_TAB:
                    signals.emit(Signal.NEW_PLOT_SETTINGS)
                # - Create plot
            elif action == Action.DELETE_RESULT:
                if context == Context.DATA_SETS_TAB:
                    signals.emit(
                        Signal.DELETE_DATA_SET,
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.KRAMERS_KRONIG_TAB:
                    signals.emit(
                        Signal.DELETE_TEST_RESULT,
                        test=project_tab.get_active_test(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.DRT_TAB:
                    signals.emit(
                        Signal.DELETE_DRT_RESULT,
                        drt=project_tab.get_active_drt(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.FITTING_TAB:
                    signals.emit(
                        Signal.DELETE_FIT_RESULT,
                        fit=project_tab.get_active_fit(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.SIMULATION_TAB:
                    signals.emit(
                        Signal.DELETE_SIMULATION_RESULT,
                        simulation=project_tab.get_active_simulation(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.PLOTTING_TAB:
                    signals.emit(
                        Signal.DELETE_PLOT_SETTINGS,
                        settings=project_tab.get_active_plot(),
                    )
                # - Plot
            elif action == Action.NEXT_PRIMARY_RESULT:
                if (
                    context == Context.DATA_SETS_TAB
                    or context == Context.KRAMERS_KRONIG_TAB
                    or context == Context.DRT_TAB
                    or context == Context.FITTING_TAB
                ):
                    signals.emit(
                        Signal.SELECT_DATA_SET,
                        data=project_tab.get_next_data_set(context),
                    )
                elif context == Context.SIMULATION_TAB:
                    signals.emit(
                        Signal.SELECT_SIMULATION_RESULT,
                        simulation=project_tab.get_active_simulation(),
                        data=project_tab.get_next_simulation_data_set(),
                    )
                elif context == Context.PLOTTING_TAB:
                    signals.emit(
                        Signal.SELECT_PLOT_SETTINGS,
                        settings=project_tab.get_next_plot(),
                    )
                # - Plot
            elif action == Action.PREVIOUS_PRIMARY_RESULT:
                if (
                    context == Context.DATA_SETS_TAB
                    or context == Context.KRAMERS_KRONIG_TAB
                    or context == Context.DRT_TAB
                    or context == Context.FITTING_TAB
                ):
                    signals.emit(
                        Signal.SELECT_DATA_SET,
                        data=project_tab.get_previous_data_set(context),
                    )
                elif context == Context.SIMULATION_TAB:
                    signals.emit(
                        Signal.SELECT_SIMULATION_RESULT,
                        simulation=project_tab.get_active_simulation(),
                        data=project_tab.get_previous_simulation_data_set(),
                    )
                elif context == Context.PLOTTING_TAB:
                    signals.emit(
                        Signal.SELECT_PLOT_SETTINGS,
                        settings=project_tab.get_previous_plot(),
                    )
                # - Plot
            elif action == Action.NEXT_SECONDARY_RESULT:
                if context == Context.KRAMERS_KRONIG_TAB:
                    signals.emit(
                        Signal.SELECT_TEST_RESULT,
                        test=project_tab.get_next_test_result(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.DRT_TAB:
                    signals.emit(
                        Signal.SELECT_DRT_RESULT,
                        drt=project_tab.get_next_drt_result(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.FITTING_TAB:
                    signals.emit(
                        Signal.SELECT_FIT_RESULT,
                        fit=project_tab.get_next_fit_result(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.SIMULATION_TAB:
                    signals.emit(
                        Signal.SELECT_SIMULATION_RESULT,
                        simulation=project_tab.get_next_simulation_result(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.PLOTTING_TAB:
                    signals.emit(
                        Signal.SELECT_PLOT_TYPE,
                        settings=project_tab.get_active_plot(),
                        plot_type=project_tab.get_next_plot_type(),
                    )
                # - Plot type
            elif action == Action.PREVIOUS_SECONDARY_RESULT:
                if context == Context.KRAMERS_KRONIG_TAB:
                    signals.emit(
                        Signal.SELECT_TEST_RESULT,
                        test=project_tab.get_previous_test_result(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.DRT_TAB:
                    signals.emit(
                        Signal.SELECT_DRT_RESULT,
                        drt=project_tab.get_previous_drt_result(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.FITTING_TAB:
                    signals.emit(
                        Signal.SELECT_FIT_RESULT,
                        fit=project_tab.get_previous_fit_result(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.SIMULATION_TAB:
                    signals.emit(
                        Signal.SELECT_SIMULATION_RESULT,
                        simulation=project_tab.get_previous_simulation_result(),
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.PLOTTING_TAB:
                    signals.emit(
                        Signal.SELECT_PLOT_TYPE,
                        settings=project_tab.get_active_plot(),
                        plot_type=project_tab.get_previous_plot_type(),
                    )
                # - Plot type
            elif action == Action.LOAD_SIMULATION_AS_DATA_SET:
                if context == Context.SIMULATION_TAB:
                    signals.emit(
                        Signal.LOAD_SIMULATION_AS_DATA_SET,
                        simulation=project_tab.get_active_simulation(),
                    )
            elif action == Action.APPLY_SETTINGS:
                if context == Context.KRAMERS_KRONIG_TAB:
                    test = project_tab.get_active_test()
                    signals.emit(
                        Signal.APPLY_TEST_SETTINGS,
                        settings=test.settings if test is not None else None,
                    )
                elif context == Context.DRT_TAB:
                    drt = project_tab.get_active_drt()
                    signals.emit(
                        Signal.APPLY_DRT_SETTINGS,
                        settings=drt.settings if drt is not None else None,
                    )
                elif context == Context.FITTING_TAB:
                    fit = project_tab.get_active_fit()
                    signals.emit(
                        Signal.APPLY_FIT_SETTINGS,
                        settings=fit.settings if fit is not None else None,
                    )
                elif context == Context.SIMULATION_TAB:
                    simulation: Optional[
                        SimulationResult
                    ] = project_tab.get_active_simulation()
                    signals.emit(
                        Signal.APPLY_SIMULATION_SETTINGS,
                        settings=simulation.settings
                        if simulation is not None
                        else None,
                    )
            elif action == Action.APPLY_MASK:
                if context == Context.KRAMERS_KRONIG_TAB:
                    test = project_tab.get_active_test()
                    signals.emit(
                        Signal.APPLY_DATA_SET_MASK,
                        mask=test.mask if test is not None else None,
                        test=test,
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.DRT_TAB:
                    drt = project_tab.get_active_drt()
                    signals.emit(
                        Signal.APPLY_DATA_SET_MASK,
                        mask=drt.mask if drt is not None else None,
                        drt=drt,
                        data=project_tab.get_active_data_set(),
                    )
                elif context == Context.FITTING_TAB:
                    fit = project_tab.get_active_fit()
                    signals.emit(
                        Signal.APPLY_DATA_SET_MASK,
                        mask=fit.mask if fit is not None else None,
                        fit=fit,
                        data=project_tab.get_active_data_set(),
                    )
            elif action == Action.SHOW_ENLARGED_DRT:
                project_tab.show_enlarged_drt()
            elif action == Action.SHOW_ENLARGED_IMPEDANCE:
                project_tab.show_enlarged_impedance()
            elif action == Action.SHOW_ENLARGED_NYQUIST:
                project_tab.show_enlarged_nyquist()
            elif action == Action.SHOW_ENLARGED_BODE:
                project_tab.show_enlarged_bode()
            elif action == Action.SHOW_ENLARGED_RESIDUALS:
                project_tab.show_enlarged_residuals()
            elif action == Action.SHOW_CIRCUIT_EDITOR:
                if context == Context.FITTING_TAB:
                    project_tab.fitting_tab.show_circuit_editor()
                elif context == Context.SIMULATION_TAB:
                    project_tab.simulation_tab.show_circuit_editor()
            elif action == Action.COPY_DRT_DATA:
                signals.emit(
                    Signal.COPY_PLOT_DATA,
                    plot=project_tab.get_drt_plot(context),
                    context=context,
                )
            elif action == Action.COPY_IMPEDANCE_DATA:
                signals.emit(
                    Signal.COPY_PLOT_DATA,
                    plot=project_tab.get_impedance_plot(context),
                    context=context,
                )
            elif action == Action.COPY_NYQUIST_DATA:
                signals.emit(
                    Signal.COPY_PLOT_DATA,
                    plot=project_tab.get_nyquist_plot(context),
                    context=context,
                )
            elif action == Action.COPY_BODE_DATA:
                signals.emit(
                    Signal.COPY_PLOT_DATA,
                    plot=project_tab.get_bode_plot(context),
                    context=context,
                )
            elif action == Action.COPY_RESIDUALS_DATA:
                signals.emit(
                    Signal.COPY_PLOT_DATA,
                    plot=project_tab.get_residuals_plot(context),
                    context=context,
                )
            elif action == Action.COPY_OUTPUT:
                if context == Context.DRT_TAB:
                    signals.emit(
                        Signal.COPY_OUTPUT,
                        output=project_tab.drt_tab.get_active_output(),
                        drt=project_tab.get_active_drt(),
                    )
                elif context == Context.FITTING_TAB:
                    signals.emit(
                        Signal.COPY_OUTPUT,
                        output=project_tab.get_active_output(context),
                        fit_or_sim=project_tab.get_active_fit(),
                    )
                elif context == Context.SIMULATION_TAB:
                    signals.emit(
                        Signal.COPY_OUTPUT,
                        output=project_tab.get_active_output(context),
                        fit_or_sim=project_tab.get_active_simulation(),
                    )
                else:
                    raise Exception(f"Unsupported context: {context=}")
            elif context == Context.DATA_SETS_TAB:
                data: Optional[DataSet] = project_tab.get_active_data_set()
                # Project-level: data sets tab
                if action == Action.AVERAGE_DATA_SETS:
                    signals.emit(
                        Signal.SELECT_DATA_SETS_TO_AVERAGE,
                    )
                elif action == Action.TOGGLE_DATA_POINTS:
                    signals.emit(
                        Signal.SELECT_DATA_POINTS_TO_TOGGLE,
                        data=data,
                    )
                elif action == Action.COPY_DATA_SET_MASK:
                    signals.emit(
                        Signal.SELECT_DATA_SET_MASK_TO_COPY,
                        data=data,
                    )
                elif action == Action.SUBTRACT_IMPEDANCE:
                    signals.emit(
                        Signal.SELECT_IMPEDANCE_TO_SUBTRACT,
                        data=data,
                    )
            elif context == Context.PLOTTING_TAB:
                settings = project_tab.get_active_plot()  # Optional[PlotSettings]
                # Project-level: plotting tab
                data_sets: List[DataSet]
                tests: List[TestResult]
                fits: List[FitResult]
                simulations: List[SimulationResult]
                if action == Action.SELECT_ALL_PLOT_SERIES:
                    (
                        data_sets,
                        tests,
                        fits,
                        simulations,
                    ) = project_tab.get_filtered_plot_series()
                    signals.emit(
                        Signal.TOGGLE_PLOT_SERIES,
                        enabled=True,
                        data_sets=data_sets,
                        tests=tests,
                        fits=fits,
                        simulations=simulations,
                        settings=settings,
                    )
                elif action == Action.UNSELECT_ALL_PLOT_SERIES:
                    (
                        data_sets,
                        tests,
                        fits,
                        simulations,
                    ) = project_tab.get_filtered_plot_series()
                    signals.emit(
                        Signal.TOGGLE_PLOT_SERIES,
                        enabled=False,
                        data_sets=data_sets,
                        tests=tests,
                        fits=fits,
                        simulations=simulations,
                        settings=settings,
                    )
                elif action == Action.COPY_PLOT_APPEARANCE:
                    signals.emit(
                        Signal.SELECT_PLOT_APPEARANCE_SETTINGS,
                        settings=settings,
                    )
                elif action == Action.COPY_PLOT_DATA:
                    signals.emit(
                        Signal.COPY_PLOT_DATA,
                        plot=project_tab.plotting_tab.plot_types[settings.get_type()]
                        if settings is not None
                        else None,
                        context=context,
                    )
                elif action == Action.EXPAND_COLLAPSE_SIDEBAR:
                    project_tab.plotting_tab.collapse_expand_sidebar()
                elif action == Action.EXPORT_PLOT:
                    signals.emit(
                        Signal.EXPORT_PLOT,
                        settings=settings,
                    )



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
)
from numpy import (
    array,
    ndarray,
)
import pyimpspec
from pyimpspec import (
    Circuit,
    Element,
    FittedParameter,
)
import dearpygui.dearpygui as dpg
from deareis.signals import Signal
import deareis.signals as signals
import deareis.themes as themes
from deareis.utility import (
    align_numbers,
    calculate_window_position_dimensions,
    format_number,
)
from deareis.tooltips import (
    attach_tooltip,
    update_tooltip,
)
import deareis.tooltips as tooltips
from deareis.gui.plots import (
    Bode,
    Nyquist,
)
from deareis.enums import (
    Context,
    FitSimOutput,
    label_to_fit_sim_output,
)
from deareis.gui.circuit_editor import (
    CircuitPreview,
    CircuitEditor,
)
from deareis.data import (
    DataSet,
    SimulationResult,
    SimulationSettings,
)


class SettingsMenu:
    def __init__(
        self,
        default_settings: SimulationSettings,
        label_pad: int,
        circuit_editor: Optional[CircuitEditor] = None,
    ):
        self.circuit_editor: Optional[CircuitEditor] = circuit_editor
        with dpg.group(horizontal=True):
            dpg.add_text("Circuit".rjust(label_pad))
            attach_tooltip(tooltips.simulation.cdc)
            self.cdc_input: int = dpg.generate_uuid()
            dpg.add_input_text(
                width=-50 if circuit_editor is not None else -1,
                tag=self.cdc_input,
                on_enter=True,
                callback=lambda s, a, u: self.parse_cdc(a, s),
            )
            self.cdc_tooltip: int = dpg.generate_uuid()
            attach_tooltip("", tag=self.cdc_tooltip, parent=self.cdc_input)
            dpg.hide_item(dpg.get_item_parent(self.cdc_tooltip))
            if circuit_editor is not None:
                self.editor_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Edit",
                    callback=self.show_circuit_editor,
                    width=-1,
                    tag=self.editor_button,
                )
                attach_tooltip(tooltips.general.open_circuit_editor)
        with dpg.group(horizontal=True):
            dpg.add_text("Maximum frequency".rjust(label_pad))
            attach_tooltip(tooltips.simulation.max_freq)
            self.max_freq_input: int = dpg.generate_uuid()
            dpg.add_input_float(
                label="Hz",
                default_value=1e5,
                min_value=1e-20,
                max_value=1e20,
                min_clamped=True,
                max_clamped=True,
                on_enter=True,
                format="%.3g",
                step=0.0,
                width=-18,
                tag=self.max_freq_input,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Minimum frequency".rjust(label_pad))
            attach_tooltip(tooltips.simulation.min_freq)
            self.min_freq_input: int = dpg.generate_uuid()
            dpg.add_input_float(
                label="Hz",
                default_value=1e-2,
                min_value=1e-20,
                max_value=1e20,
                min_clamped=True,
                max_clamped=True,
                on_enter=True,
                format="%.3g",
                step=0.0,
                width=-18,
                tag=self.min_freq_input,
            )
        with dpg.group(horizontal=True):
            dpg.add_text("Num. points per decade".rjust(label_pad))
            attach_tooltip(tooltips.simulation.per_decade)
            self.per_decade_input: int = dpg.generate_uuid()
            dpg.add_input_int(
                default_value=10,
                min_value=1,
                min_clamped=True,
                step=0,
                on_enter=True,
                tag=self.per_decade_input,
                width=-1,
            )
        self.set_settings(default_settings)

    def get_settings(self) -> SimulationSettings:
        cdc: str = dpg.get_value(self.cdc_input) or ""
        circuit: Optional[Circuit] = dpg.get_item_user_data(self.cdc_input)
        if circuit is None or cdc != circuit.to_string():
            circuit = self.parse_cdc(cdc, self.cdc_input)
        min_f: float = dpg.get_value(self.min_freq_input)
        max_f: float = dpg.get_value(self.max_freq_input)
        if min_f > max_f:
            max_f = min_f
            min_f = dpg.get_value(self.max_freq_input)
            dpg.set_value(self.min_freq_input, min_f)
            dpg.set_value(self.max_freq_input, max_f)
        return SimulationSettings(
            cdc=circuit.to_string(12) if circuit is not None else "",
            min_frequency=min_f,
            max_frequency=max_f,
            num_per_decade=dpg.get_value(self.per_decade_input),
        )

    def set_settings(self, settings: SimulationSettings):
        assert type(settings) is SimulationSettings, settings
        self.parse_cdc(settings.cdc)
        dpg.set_value(self.min_freq_input, settings.min_frequency)
        dpg.set_value(self.max_freq_input, settings.max_frequency)
        dpg.set_value(self.per_decade_input, settings.num_per_decade)

    def parse_cdc(self, cdc: str, sender: int = -1) -> Optional[Circuit]:
        assert type(cdc) is str, cdc
        assert type(sender) is int, sender
        try:
            circuit: Circuit = pyimpspec.parse_cdc(cdc)
        except (pyimpspec.ParsingError, pyimpspec.UnexpectedCharacter) as err:
            dpg.bind_item_theme(self.cdc_input, themes.cdc.invalid)
            update_tooltip(self.cdc_tooltip, str(err))
            dpg.show_item(dpg.get_item_parent(self.cdc_tooltip))
            dpg.set_item_user_data(self.cdc_input, None)
            return None
        dpg.bind_item_theme(self.cdc_input, themes.cdc.valid)
        dpg.hide_item(dpg.get_item_parent(self.cdc_tooltip))
        dpg.set_item_user_data(self.cdc_input, circuit)
        if sender != self.cdc_input:
            dpg.set_value(self.cdc_input, circuit.to_string())
        return circuit

    def show_circuit_editor(self):
        if self.circuit_editor is None:
            return
        x: int
        y: int
        w: int
        h: int
        x, y, w, h = calculate_window_position_dimensions()
        dpg.configure_item(
            self.circuit_editor.window,
            pos=(
                x,
                y,
            ),
            width=w,
            height=h,
        )
        circuit: Optional[Circuit] = None
        try:
            circuit = pyimpspec.parse_cdc(self.get_settings().cdc)
        except pyimpspec.ParsingError:
            pass
        signals.emit(
            Signal.BLOCK_KEYBINDINGS,
            window=self.circuit_editor.window,
            window_object=self.circuit_editor,
        )
        self.circuit_editor.show(circuit)

    def has_active_input(self) -> bool:
        return (
            dpg.is_item_active(self.cdc_input)
            or dpg.is_item_active(self.max_freq_input)
            or dpg.is_item_active(self.min_freq_input)
            or dpg.is_item_active(self.per_decade_input)
        )


class ParametersTable:
    def __init__(self):
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(label=" Parameters", leaf=True, tag=self._header):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Element",
                    width_fixed=True,
                )
                attach_tooltip(tooltips.simulation.element)
                dpg.add_table_column(
                    label="Parameter",
                    width_fixed=True,
                )
                attach_tooltip(tooltips.simulation.parameter)
                dpg.add_table_column(
                    label="Value",
                    width_fixed=True,
                )
                attach_tooltip(tooltips.simulation.parameter_value)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        dpg.delete_item(self._table, children_only=True, slot=1)

    def populate(self, simulation: SimulationResult):
        dpg.show_item(self._header)
        column_pads: List[int] = [
            10,
            11,
            10,
        ]
        if simulation is None:
            with dpg.table_row(parent=self._table):
                dpg.add_text("".ljust(column_pads[0]))
                dpg.add_text("".ljust(column_pads[1]))
                dpg.add_text("".ljust(column_pads[2]))
            return
        element_labels: List[str] = []
        element_tooltips: List[str] = []
        parameter_labels: List[str] = []
        values: List[str] = []
        value_tooltips: List[str] = []
        element_label: str
        element_tooltip: str
        parameter_label: str
        value: str
        value_tooltip: str
        parameters: Dict[str, FittedParameter]
        for element in simulation.circuit.get_elements():
            float_value: float
            for parameter_label, float_value in element.get_parameters().items():
                element_labels.append(element.get_label())
                element_tooltips.append(element.get_extended_description())
                parameter_labels.append(parameter_label)
                values.append(f"{format_number(float_value, width=9, significants=3)}")
                value_tooltips.append(
                    f"{format_number(float_value, decimals=6).strip()}"
                )
        values = align_numbers(values)
        num_rows: int = 0
        for (
            element_label,
            element_tooltip,
            parameter_label,
            value,
            value_tooltip,
        ) in zip(
            element_labels,
            element_tooltips,
            parameter_labels,
            values,
            value_tooltips,
        ):
            with dpg.table_row(parent=self._table):
                dpg.add_text(element_label.ljust(column_pads[0]))
                if element_tooltip != "":
                    attach_tooltip(element_tooltip)
                dpg.add_text(parameter_label.ljust(column_pads[1]))
                dpg.add_text(value.ljust(column_pads[2]))
                attach_tooltip(value_tooltip)
                num_rows += 1
        dpg.set_item_height(self._table, 18 + 23 * max(1, num_rows))


class SettingsTable:
    def __init__(self):
        label_pad: int = 23
        self._header: int = dpg.generate_uuid()
        with dpg.collapsing_header(label=" Settings", leaf=True, tag=self._header):
            self._table: int = dpg.generate_uuid()
            with dpg.table(
                borders_outerV=True,
                borders_outerH=True,
                borders_innerV=True,
                borders_innerH=True,
                scrollY=True,
                freeze_rows=1,
                height=18 + 23,
                tag=self._table,
            ):
                dpg.add_table_column(
                    label="Label".rjust(label_pad),
                    width_fixed=True,
                )
                dpg.add_table_column(
                    label="Value",
                    width_fixed=True,
                )
                label: str
                for label in [
                    "Circuit",
                    "Maximum frequency",
                    "Minimum frequency",
                    "Points per decade",
                ]:
                    with dpg.table_row():
                        dpg.add_text(label.rjust(label_pad))
                        tooltip_tag: int = dpg.generate_uuid()
                        dpg.add_text("", user_data=tooltip_tag)
                        attach_tooltip("", tag=tooltip_tag)
            with dpg.group(horizontal=True):
                self._apply_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Apply settings",
                    callback=lambda s, a, u: signals.emit(
                        Signal.APPLY_SIMULATION_SETTINGS,
                        **u,
                    ),
                    tag=self._apply_button,
                )
                attach_tooltip(tooltips.general.apply_settings)
                self._load_as_data_button: int = dpg.generate_uuid()
                dpg.add_button(
                    label="Load as data set",
                    callback=lambda s, a, u: signals.emit(
                        Signal.LOAD_SIMULATION_AS_DATA_SET,
                        **u,
                    ),
                    tag=self._load_as_data_button,
                )
                attach_tooltip(tooltips.simulation.load_as_data_set)

    def clear(self, hide: bool):
        if hide:
            dpg.hide_item(self._header)
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            tag: int = dpg.get_item_children(row, slot=1)[1]
            dpg.set_value(tag, "")
            dpg.hide_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))

    def populate(self, simulation: SimulationResult):
        settings: SimulationSettings = simulation.settings
        dpg.show_item(self._header)
        rows: List[int] = []
        cells: List[Tuple[int, int]] = []
        row: int
        for row in dpg.get_item_children(self._table, slot=1):
            rows.append(row)
            cells.append(dpg.get_item_children(row, slot=1))
        assert len(rows) == len(cells) == 4, (
            rows,
            cells,
        )
        circuit: Circuit = pyimpspec.parse_cdc(settings.cdc)
        tag: int
        value: str
        for (row, tag, value) in [
            (
                rows[0],
                cells[0][1],
                f"{circuit}\n\n{circuit.to_string(3)}",
            ),
            (
                rows[1],
                cells[1][1],
                format_number(settings.max_frequency, 3, 12),
            ),
            (
                rows[2],
                cells[2][1],
                format_number(settings.min_frequency, 3, 12),
            ),
            (
                rows[3],
                cells[3][1],
                f"{settings.num_per_decade}",
            ),
        ]:
            dpg.set_value(tag, value.split("\n")[0])
            update_tooltip(dpg.get_item_user_data(tag), value)
            dpg.show_item(dpg.get_item_parent(dpg.get_item_user_data(tag)))
        dpg.set_item_height(self._table, 18 + 23 * 4)
        dpg.set_item_user_data(
            self._apply_button,
            {"settings": settings},
        )
        dpg.set_item_user_data(
            self._load_as_data_button,
            {"simulation": simulation},
        )


class DataSetsCombo:
    def __init__(self, label: str, width: int, callback: Callable):
        self.labels: List[str] = []
        dpg.add_text(label)
        self.tag: int = dpg.generate_uuid()
        dpg.add_combo(
            callback=callback,
            user_data={},
            width=width,
            tag=self.tag,
        )

    def populate(self, labels: List[str], lookup: Dict[str, DataSet]):
        self.labels.clear()
        self.labels.append("")
        self.labels.extend(labels)
        label: str = dpg.get_value(self.tag) or ""
        if labels and label not in labels:
            label = ""
        dpg.configure_item(
            self.tag,
            default_value=label,
            items=[""] + labels,
            user_data=lookup,
        )

    def get(self) -> Optional[DataSet]:
        return dpg.get_item_user_data(self.tag).get(dpg.get_value(self.tag))

    def set(self, label: str):
        assert type(label) is str, label
        assert label in self.labels
        dpg.set_value(self.tag, label)

    def clear(self):
        dpg.configure_item(
            self.tag,
            default_value="",
        )

    def get_next(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        current_label: str = dpg.get_value(self.tag)
        if current_label not in labels:
            return lookup[labels[0]]
        index: int = labels.index(current_label) + 1
        if index == len(labels):
            return None
        return lookup[labels[index % len(labels)]]

    def get_previous(self) -> Optional[DataSet]:
        lookup: Dict[str, DataSet] = dpg.get_item_user_data(self.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        current_label: str = dpg.get_value(self.tag)
        if current_label not in labels:
            return lookup[labels[-1]]
        index: int = labels.index(current_label) - 1
        if index == -1:
            return None
        return lookup[labels[index % len(labels)]]


class ResultsCombo:
    def __init__(self, label: str, width: int, callback: Callable):
        self.labels: Dict[str, str] = {}
        dpg.add_text(label)
        self.tag: int = dpg.generate_uuid()
        dpg.add_combo(
            callback=callback,
            user_data={},
            width=width,
            tag=self.tag,
        )

    def populate(self, lookup: Dict[str, SimulationResult]):
        self.labels.clear()
        labels: List[str] = list(lookup.keys())
        longest_cdc: int = max(list(map(lambda _: len(_[: _.find(" ")]), labels)) + [1])
        old_key: str
        for old_key in labels:
            simulation: SimulationResult = lookup[old_key]
            del lookup[old_key]
            cdc, timestamp = (
                old_key[: old_key.find(" ")],
                old_key[old_key.find(" ") + 1 :],
            )
            new_key: str = f"{cdc.ljust(longest_cdc)} {timestamp}"
            self.labels[old_key] = new_key
            lookup[new_key] = simulation
        labels = list(lookup.keys())
        dpg.configure_item(
            self.tag,
            default_value=labels[0] if labels else "",
            items=labels,
            user_data=lookup,
        )

    def get(self) -> Optional[SimulationResult]:
        return dpg.get_item_user_data(self.tag).get(dpg.get_value(self.tag))

    def set(self, label: str):
        assert type(label) is str, label
        assert label in self.labels, (
            label,
            list(self.labels.keys()),
        )
        dpg.set_value(self.tag, self.labels[label])

    def clear(self):
        dpg.configure_item(
            self.tag,
            default_value="",
        )

    def get_next(self) -> Optional[SimulationResult]:
        lookup: Dict[str, SimulationResult] = dpg.get_item_user_data(self.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) + 1
        return lookup[labels[index % len(labels)]]

    def get_previous(self) -> Optional[SimulationResult]:
        lookup: Dict[str, SimulationResult] = dpg.get_item_user_data(self.tag)
        if not lookup:
            return None
        labels: List[str] = list(lookup.keys())
        index: int = labels.index(dpg.get_value(self.tag)) - 1
        return lookup[labels[index % len(labels)]]


class SimulationTab:
    def __init__(self, state):
        self.state = state
        self.queued_update: Optional[Callable] = None
        self.tab: int = dpg.generate_uuid()
        with dpg.tab(label="Simulation", tag=self.tab):
            self.sidebar_width: int = 350
            settings_height: int = 128
            label_pad: int = 24
            with dpg.group(horizontal=True):
                self.sidebar_window: int = dpg.generate_uuid()
                with dpg.child_window(
                    border=False,
                    width=self.sidebar_width,
                    tag=self.sidebar_window,
                ):
                    # Settings
                    with dpg.child_window(
                        border=True, width=-1, height=settings_height
                    ):
                        self.circuit_editor: CircuitEditor = CircuitEditor(
                            window=dpg.add_window(
                                label="Circuit editor",
                                show=False,
                                modal=True,
                                on_close=lambda s, a, u: self.accept_circuit(None),
                            ),
                            callback=self.accept_circuit,
                        )
                        self.settings_menu: SettingsMenu = SettingsMenu(
                            state.config.default_simulation_settings,
                            label_pad,
                            circuit_editor=self.circuit_editor,
                        )
                        with dpg.group(horizontal=True):
                            self.visibility_item: int = dpg.generate_uuid()
                            dpg.add_text("".rjust(label_pad), tag=self.visibility_item)
                            self.perform_sim_button: int = dpg.generate_uuid()
                            dpg.add_button(
                                label="Perform simulation",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.PERFORM_SIMULATION,
                                    data=u,
                                    settings=self.get_settings(),
                                ),
                                user_data=None,
                                width=-1,
                                tag=self.perform_sim_button,
                            )
                    with dpg.child_window(width=-1, height=82):
                        label_pad = 8
                        with dpg.group(horizontal=True):
                            self.data_sets_combo: DataSetsCombo = DataSetsCombo(
                                label="Data set".rjust(label_pad),
                                width=-60,
                                callback=lambda s, a, u: signals.emit(
                                    Signal.SELECT_SIMULATION_RESULT,
                                    simulation=self.results_combo.get(),
                                    data=u.get(a),
                                ),
                            )
                        with dpg.group(horizontal=True):
                            self.results_combo: ResultsCombo = ResultsCombo(
                                label="Result".rjust(label_pad),
                                width=-60,
                                callback=lambda s, a, u: signals.emit(
                                    Signal.SELECT_SIMULATION_RESULT,
                                    data=self.data_sets_combo.get(),
                                    simulation=u.get(a),
                                ),
                            )
                            self.delete_button: int = dpg.generate_uuid()
                            dpg.add_button(
                                label="Delete",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.DELETE_SIMULATION_RESULT, simulation=u
                                ),
                                width=-1,
                                tag=self.delete_button,
                            )
                            attach_tooltip(tooltips.simulation.remove)
                        with dpg.group(horizontal=True):
                            dpg.add_text("Output".rjust(label_pad))
                            # TODO: Split into combo class?
                            self.output_combo: int = dpg.generate_uuid()
                            dpg.add_combo(
                                items=list(label_to_fit_sim_output.keys()),
                                default_value=list(label_to_fit_sim_output.keys())[0],
                                tag=self.output_combo,
                                width=-60,
                            )
                            self.copy_output_button: int = dpg.generate_uuid()
                            dpg.add_button(
                                label="Copy",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.COPY_OUTPUT,
                                    output=self.get_active_output(),
                                    **u,
                                ),
                                user_data={},
                                width=-1,
                                tag=self.copy_output_button,
                            )
                            attach_tooltip(tooltips.general.copy_output)
                    with dpg.child_window(width=-1, height=-1):
                        self.result_group: int = dpg.generate_uuid()
                        with dpg.group(tag=self.result_group):
                            self.parameters_table: ParametersTable = ParametersTable()
                            dpg.add_spacer(height=8)
                            self.settings_table: SettingsTable = SettingsTable()
                self.plot_window: int = dpg.generate_uuid()
                with dpg.child_window(
                    border=False,
                    width=-1,
                    height=-1,
                    tag=self.plot_window,
                ):
                    self.minimum_plot_side: int = 400
                    with dpg.group(horizontal=True):
                        self.circuit_preview_height: int = 250
                        with dpg.child_window(
                            border=False,
                            width=-1,
                            height=self.circuit_preview_height,
                        ):
                            dpg.add_text("Simulated circuit")
                            self.circuit_preview: CircuitPreview = CircuitPreview()
                    with dpg.group(horizontal=True):
                        with dpg.group():
                            self.nyquist_plot: Nyquist = Nyquist(
                                width=self.minimum_plot_side,
                                height=self.minimum_plot_side,
                            )
                            self.nyquist_plot.plot(
                                real=array([]),
                                imaginary=array([]),
                                label="Data",
                                theme=themes.nyquist.data,
                            )
                            self.nyquist_plot.plot(
                                real=array([]),
                                imaginary=array([]),
                                label="Sim.",
                                simulation=True,
                                theme=themes.nyquist.simulation,
                            )
                            self.nyquist_plot.plot(
                                real=array([]),
                                imaginary=array([]),
                                label="Sim.",
                                simulation=True,
                                line=True,
                                theme=themes.nyquist.simulation,
                                show_label=False,
                            )
                            with dpg.group(horizontal=True):
                                self.enlarge_nyquist_button: int = dpg.generate_uuid()
                                self.adjust_nyquist_limits_checkbox: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Enlarge Nyquist",
                                    callback=self.show_enlarged_nyquist,
                                    tag=self.enlarge_nyquist_button,
                                )
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_nyquist_limits_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_nyquist_limits)
                                dpg.add_button(
                                    label="Copy as CSV",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.COPY_PLOT_DATA,
                                        plot=self.nyquist_plot,
                                        context=Context.FITTING_TAB,
                                    ),
                                )
                                attach_tooltip(tooltips.general.copy_plot_data_as_csv)
                        self.horizontal_bode_group: int = dpg.generate_uuid()
                        with dpg.group(tag=self.horizontal_bode_group):
                            self.bode_plot_horizontal: Bode = Bode(
                                width=self.minimum_plot_side,
                                height=self.minimum_plot_side,
                            )
                            self.bode_plot_horizontal.plot(
                                frequency=array([]),
                                magnitude=array([]),
                                phase=array([]),
                                labels=(
                                    "|Z| (d)",
                                    "phi (d)",
                                ),
                                themes=(
                                    themes.bode.magnitude_data,
                                    themes.bode.phase_data,
                                ),
                            )
                            self.bode_plot_horizontal.plot(
                                frequency=array([]),
                                magnitude=array([]),
                                phase=array([]),
                                labels=(
                                    "|Z| (s)",
                                    "phi (s)",
                                ),
                                simulation=True,
                                themes=(
                                    themes.bode.magnitude_simulation,
                                    themes.bode.phase_simulation,
                                ),
                            )
                            self.bode_plot_horizontal.plot(
                                frequency=array([]),
                                magnitude=array([]),
                                phase=array([]),
                                labels=(
                                    "|Z| (s)",
                                    "phi (s)",
                                ),
                                simulation=True,
                                line=True,
                                themes=(
                                    themes.bode.magnitude_simulation,
                                    themes.bode.phase_simulation,
                                ),
                                show_labels=False,
                            )
                            with dpg.group(horizontal=True):
                                self.enlarge_bode_horizontal_button: int = (
                                    dpg.generate_uuid()
                                )
                                self.adjust_bode_limits_horizontal_checkbox: int = (
                                    dpg.generate_uuid()
                                )
                                dpg.add_button(
                                    label="Enlarge Bode",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.SHOW_ENLARGED_PLOT,
                                        plot=self.bode_plot_horizontal,
                                        adjust_limits=dpg.get_value(
                                            self.adjust_bode_limits_horizontal_checkbox
                                        ),
                                    ),
                                    tag=self.enlarge_bode_horizontal_button,
                                )
                                dpg.add_checkbox(
                                    default_value=True,
                                    tag=self.adjust_bode_limits_horizontal_checkbox,
                                )
                                attach_tooltip(tooltips.general.adjust_bode_limits)
                                dpg.add_button(
                                    label="Copy as CSV",
                                    callback=lambda s, a, u: signals.emit(
                                        Signal.COPY_PLOT_DATA,
                                        plot=self.bode_plot_horizontal,
                                        context=Context.FITTING_TAB,
                                    ),
                                )
                                attach_tooltip(tooltips.general.copy_plot_data_as_csv)
                    self.vertical_bode_group: int = dpg.generate_uuid()
                    with dpg.group(tag=self.vertical_bode_group, show=False):
                        self.bode_plot_vertical: Bode = Bode(
                            width=self.minimum_plot_side, height=self.minimum_plot_side
                        )
                        self.bode_plot_vertical.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z| (d)",
                                "phi (d)",
                            ),
                            themes=(
                                themes.bode.magnitude_data,
                                themes.bode.phase_data,
                            ),
                        )
                        self.bode_plot_vertical.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z| (s)",
                                "phi (s)",
                            ),
                            simulation=True,
                            themes=(
                                themes.bode.magnitude_simulation,
                                themes.bode.phase_simulation,
                            ),
                        )
                        self.bode_plot_vertical.plot(
                            frequency=array([]),
                            magnitude=array([]),
                            phase=array([]),
                            labels=(
                                "|Z| (s)",
                                "phi (s)",
                            ),
                            simulation=True,
                            line=True,
                            themes=(
                                themes.bode.magnitude_simulation,
                                themes.bode.phase_simulation,
                            ),
                            show_labels=False,
                        )
                        with dpg.group(horizontal=True):
                            self.enlarge_bode_vertical_button: int = dpg.generate_uuid()
                            self.adjust_bode_limits_vertical_checkbox: int = (
                                dpg.generate_uuid()
                            )
                            dpg.add_button(
                                label="Enlarge Bode",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.SHOW_ENLARGED_PLOT,
                                    plot=self.bode_plot_vertical,
                                    adjust_limits=dpg.get_value(
                                        self.adjust_bode_limits_horizontal_checkbox
                                    ),
                                ),
                                tag=self.enlarge_bode_vertical_button,
                            )
                            dpg.add_checkbox(
                                default_value=True,
                                source=self.adjust_bode_limits_horizontal_checkbox,
                                tag=self.adjust_bode_limits_vertical_checkbox,
                            )
                            attach_tooltip(tooltips.general.adjust_bode_limits)
                            dpg.add_button(
                                label="Copy as CSV",
                                callback=lambda s, a, u: signals.emit(
                                    Signal.COPY_PLOT_DATA,
                                    plot=self.bode_plot_vertical,
                                    context=Context.FITTING_TAB,
                                ),
                            )
                            attach_tooltip(tooltips.general.copy_plot_data_as_csv)
        self.set_settings(self.state.config.default_simulation_settings)

    def is_visible(self) -> bool:
        return dpg.is_item_visible(self.visibility_item)

    def get_settings(self) -> SimulationSettings:
        return self.settings_menu.get_settings()

    def set_settings(self, settings: SimulationSettings):
        self.settings_menu.set_settings(settings)

    def resize(self, width: int, height: int):
        assert type(width) is int and width > 0
        assert type(height) is int and height > 0
        if not self.is_visible():
            return
        if width < (self.sidebar_width + self.minimum_plot_side * 2):
            if dpg.is_item_shown(self.horizontal_bode_group):
                dpg.hide_item(self.horizontal_bode_group)
                dpg.show_item(self.vertical_bode_group)
                self.nyquist_plot.resize(-1, self.minimum_plot_side)
                self.bode_plot_vertical.resize(-1, self.minimum_plot_side)
        else:
            if dpg.is_item_shown(self.vertical_bode_group):
                dpg.show_item(self.horizontal_bode_group)
                dpg.hide_item(self.vertical_bode_group)
            dpg.split_frame()
            width, height = dpg.get_item_rect_size(self.plot_window)
            width = round((width - 8) / 2)
            height = height - 228 - 24 * 2 - 2
            self.nyquist_plot.resize(width, height)
            self.bode_plot_horizontal.resize(width, height)

    def clear(self, hide: bool = True):
        self.data_sets_combo.clear()
        self.results_combo.clear()
        self.parameters_table.clear(hide=hide)
        self.settings_table.clear(hide=hide)
        self.circuit_preview.clear()
        self.nyquist_plot.clear(delete=False)
        self.bode_plot_horizontal.clear(delete=False)
        self.bode_plot_vertical.clear(delete=False)

    def populate_data_sets(self, labels: List[str], lookup: Dict[str, DataSet]):
        assert type(labels) is list, labels
        assert type(lookup) is dict, lookup
        self.data_sets_combo.populate(labels, lookup)

    def populate_simulations(self, lookup: Dict[str, SimulationResult]):
        assert type(lookup) is dict, lookup
        if not lookup:
            self.clear()
        self.results_combo.populate(lookup)
        data: Optional[DataSet] = dpg.get_item_user_data(self.perform_sim_button)
        if not self.results_combo.labels:
            self.select_simulation_result(None, data)
            return
        signals.emit(
            Signal.SELECT_SIMULATION_RESULT,
            simulation=self.results_combo.get(),
            data=data,
        )

    def get_next_data_set(self) -> Optional[DataSet]:
        return self.data_sets_combo.get_next()

    def get_previous_data_set(self) -> Optional[DataSet]:
        return self.data_sets_combo.get_previous()

    def get_next_result(self) -> Optional[SimulationResult]:
        return self.results_combo.get_next()

    def get_previous_result(self) -> Optional[SimulationResult]:
        return self.results_combo.get_previous()

    def select_data_set(self, data: Optional[DataSet]):
        assert type(data) is DataSet or data is None, data
        self.clear()
        dpg.set_item_user_data(self.perform_sim_button, data)
        if data is None:
            return
        self.data_sets_combo.set(data.get_label())
        real: ndarray
        imag: ndarray
        real, imag = data.get_nyquist_data()
        self.nyquist_plot.update(
            index=0,
            real=real,
            imaginary=imag,
        )
        freq: ndarray
        mag: ndarray
        phase: ndarray
        freq, mag, phase = data.get_bode_data()
        self.bode_plot_horizontal.update(
            index=0,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        self.bode_plot_vertical.update(
            index=0,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )

    def select_simulation_result(
        self, simulation: Optional[SimulationResult], data: Optional[DataSet]
    ):
        assert type(simulation) is SimulationResult or simulation is None, simulation
        assert type(data) is DataSet or data is None, data
        dpg.set_item_user_data(
            self.delete_button,
            simulation,
        )
        dpg.set_item_user_data(
            self.copy_output_button,
            {
                "fit_or_sim": simulation,
                "data": data,
            },
        )
        if not self.is_visible():
            self.queued_update = lambda: self.select_simulation_result(simulation, data)
            return
        self.queued_update = None
        self.select_data_set(data)
        if simulation is None:
            if dpg.get_value(self.adjust_nyquist_limits_checkbox):
                self.nyquist_plot.queue_limits_adjustment()
            if dpg.get_value(self.adjust_bode_limits_horizontal_checkbox):
                self.bode_plot_horizontal.queue_limits_adjustment()
                self.bode_plot_vertical.queue_limits_adjustment()
            return
        self.results_combo.set(simulation.get_label())
        self.parameters_table.populate(simulation)
        self.settings_table.populate(simulation)
        self.circuit_preview.update(pyimpspec.parse_cdc(simulation.settings.cdc))
        real: ndarray
        imag: ndarray
        real, imag = simulation.get_nyquist_data()
        self.nyquist_plot.update(
            index=1,
            real=real,
            imaginary=imag,
        )
        real, imag = simulation.get_nyquist_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.nyquist_plot.update(
            index=2,
            real=real,
            imaginary=imag,
        )
        freq: ndarray
        mag: ndarray
        phase: ndarray
        freq, mag, phase = simulation.get_bode_data()
        self.bode_plot_horizontal.update(
            index=1,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = simulation.get_bode_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.bode_plot_horizontal.update(
            index=2,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = simulation.get_bode_data()
        self.bode_plot_vertical.update(
            index=1,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        freq, mag, phase = simulation.get_bode_data(
            num_per_decade=self.state.config.num_per_decade_in_simulated_lines
        )
        self.bode_plot_vertical.update(
            index=2,
            frequency=freq,
            magnitude=mag,
            phase=phase,
        )
        if dpg.get_value(self.adjust_nyquist_limits_checkbox):
            self.nyquist_plot.queue_limits_adjustment()
        if dpg.get_value(self.adjust_bode_limits_horizontal_checkbox):
            self.bode_plot_horizontal.queue_limits_adjustment()
            self.bode_plot_vertical.queue_limits_adjustment()

    def show_circuit_editor(self):
        self.settings_menu.show_circuit_editor()

    def accept_circuit(self, circuit: Optional[Circuit]):
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)
        self.circuit_editor.hide()
        if circuit is None:
            return
        self.settings_menu.parse_cdc(circuit.to_string(12))

    def show_enlarged_nyquist(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.nyquist_plot,
            adjust_limits=dpg.get_value(self.adjust_nyquist_limits_checkbox),
        )

    def show_enlarged_bode(self):
        signals.emit(
            Signal.SHOW_ENLARGED_PLOT,
            plot=self.bode_plot_horizontal,
            adjust_limits=dpg.get_value(self.adjust_bode_limits_horizontal_checkbox),
        )

    def get_active_output(self) -> Optional[FitSimOutput]:
        return label_to_fit_sim_output.get(dpg.get_value(self.output_combo))

    def has_active_input(self) -> bool:
        return self.settings_menu.has_active_input()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from types import SimpleNamespace


home = SimpleNamespace(
    **{
        "load": """
Load this project.
    """.strip(),
        "new_project": """
Create a new project.
    """.strip(),
        "load_projects": """
Select project(s) to load. If one or more recent projects are marked, then those will be loaded instead.
    """.strip(),
        "merge_projects": """
Select project(s) to merge into a new project. If one or more recent projects are marked, then those will be merged instead.

If a project file has been copied and renamed, then this function can be used to reassign all of the universally uniqued identifiers (UUIDs) in the copy. This will then prevent DearEIS from mistaking the copy for the original if there is an attempt to load the copy when the original is already open (or vice versa).
    """.strip(),
        "clear_recent_projects": """
Clear the list of recent projects. If one or more recent projects are marked, then only those will be cleared from the list.
    """.strip(),
    }
)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

import dearpygui.dearpygui as dpg
from deareis.utility import calculate_window_position_dimensions
from os import walk
from os.path import abspath, exists, dirname, join
from typing import Dict, IO, List
from deareis.signals import Signal
import deareis.signals as signals


def read_file(path: str) -> str:
    assert type(path) is str
    assert exists(path), path
    fp: IO
    with open(path, "r") as fp:
        return fp.read()


def get_licenses(root: str) -> Dict[str, str]:
    assert type(root) is str
    assert exists(root), root
    files: List[str] = []
    for _, _, files in walk(root):
        break
    files = list(filter(lambda _: not _.endswith(".py"), files))
    assert len(files) > 0
    prefix: str = "LICENSE-"
    extension: str = ".txt"
    assert all(map(lambda _: _.startswith(prefix) and _.endswith(extension), files))
    files.sort()
    licenses: Dict[str, str] = {}
    file: str
    for file in files:
        key: str = file[len(prefix) : file.rfind(extension)]
        licenses[key] = read_file(join(root, file))
    return licenses


def show_license_window():
    licenses: Dict[str, str] = get_licenses(dirname(abspath(__file__)))
    x: int
    y: int
    w: int
    h: int
    x, y, w, h = calculate_window_position_dimensions(640, 540)
    window: int = dpg.generate_uuid()
    key_handler: int = dpg.generate_uuid()

    def close_window():
        if dpg.does_item_exist(window):
            dpg.delete_item(window)
        if dpg.does_item_exist(key_handler):
            dpg.delete_item(key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    with dpg.handler_registry(tag=key_handler):
        dpg.add_key_release_handler(
            key=dpg.mvKey_Escape,
            callback=close_window,
        )

    with dpg.window(
        label="Licenses",
        modal=True,
        pos=(
            x,
            y,
        ),
        width=w,
        height=h,
        no_move=False,
        no_resize=True,
        on_close=close_window,
        tag=window,
    ):
        with dpg.tab_bar():
            with dpg.tab(label="DearEIS"):
                with dpg.child_window(border=False):
                    dpg.add_text(licenses["DearEIS"], wrap=w)
                del licenses["DearEIS"]
            with dpg.tab(label="Dependencies"):
                text_widget: int = dpg.generate_uuid()

                def show_dependency_license(sender: int, label: str):
                    dpg.set_value(text_widget, licenses[label])

                items: List[str] = list(sorted(licenses.keys()))
                dpg.add_combo(
                    items=items,
                    default_value=items[0],
                    width=-1,
                    callback=show_dependency_license,
                )
                with dpg.child_window(border=False):
                    dpg.add_text(licenses[items[0]], wrap=w, tag=text_widget)
    signals.emit(Signal.BLOCK_KEYBINDINGS, window=window, window_object=None)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from threading import Timer
from typing import (
    List,
    Optional,
)
from deareis.data import Project
from deareis.gui import ProjectTab
from deareis.signals import Signal
from deareis.state import STATE
import deareis.signals as signals


def rename_project(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    label: Optional[str] = kwargs.get("label")
    if label is None:
        return
    label = label.strip()
    if label == project.get_label():
        return
    elif label == "":
        project_tab.set_label(project.get_label())
        return
    project.set_label(label)
    project_tab.set_label(label)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)


def modify_project_notes(*args, **kwargs):
    timers: Optional[List[Timer]] = kwargs.get("timers")
    assert timers is not None, timers
    while timers:
        timers.pop(0).cancel()
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return

    def create_snapshot():
        project.set_notes(project_tab.get_notes())
        signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)

    t: Timer = Timer(0.5, create_snapshot)
    timers.append(t)
    t.start()



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    List,
    Optional,
)
import dearpygui.dearpygui as dpg
from numpy import (
    ceil,
    floor,
    integer,
    issubdtype,
    ndarray,
)
import deareis.themes as themes
from deareis.gui.plots.base import Plot


class MuXps(Plot):
    def __init__(self, width: int = -1, height: int = -1, *args, **kwargs):
        assert type(width) is int, width
        assert type(height) is int, height
        super().__init__()
        with dpg.plot(
            anti_aliased=True,
            crosshairs=True,
            width=width,
            height=height,
            tag=self._plot,
        ):
            dpg.add_plot_legend(
                horizontal=True,
                location=dpg.mvPlot_Location_North,
                outside=kwargs.get("legend_outside", True),
            )
            self._x_axis: int = dpg.add_plot_axis(
                dpg.mvXAxis,
                label="number of RC elements",
            )
            self._y_axis_1: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label="µ",
                no_gridlines=True,
            )
            self._y_axis_2: int = dpg.add_plot_axis(
                dpg.mvYAxis,
                label="log X² (pseudo)",
                no_gridlines=True,
            )
        dpg.bind_item_theme(self._plot, themes.plot)
        dpg.bind_item_handler_registry(self._plot, self._item_handler)

    @classmethod
    def duplicate(Class, original: Plot, *args, **kwargs) -> Plot:
        copy: Plot = Class(*args, **kwargs)
        for kwargs in original.get_series():
            copy.plot(**kwargs)
        return copy

    def is_blank(self) -> bool:
        return (
            len(dpg.get_item_children(self._y_axis_1, slot=1)) == 0
            and len(dpg.get_item_children(self._y_axis_2, slot=1)) == 0
        )

    def clear(self, *args, **kwargs):
        dpg.delete_item(self._y_axis_1, children_only=True)
        dpg.delete_item(self._y_axis_2, children_only=True)
        self._series.clear()

    def update(self, num_RC: int):
        assert type(num_RC) is int and num_RC >= 0, num_RC
        assert len(self._series) == 1, len(self._series)
        num_RCs: ndarray = self._series[0]["num_RCs"]
        mu: ndarray = self._series[0]["mu"]
        Xps: ndarray = self._series[0]["Xps"]
        x: list = list(num_RCs)
        y: list = list(mu)
        dpg.set_value(
            dpg.get_item_children(self._y_axis_1, slot=1)[1],
            [[num_RC], y[x.index(num_RC)]],
        )
        y = list(Xps)
        dpg.set_value(
            dpg.get_item_children(self._y_axis_2, slot=1)[1],
            [[num_RC], y[x.index(num_RC)]],
        )

    def plot(self, *args, **kwargs):
        assert len(args) == 0, args
        num_RCs: ndarray = kwargs["num_RCs"]
        mu: ndarray = kwargs["mu"]
        Xps: ndarray = kwargs["Xps"]
        mu_criterion: float = kwargs["mu_criterion"]
        num_RC: int = kwargs["num_RC"]
        assert type(num_RCs) is ndarray, num_RCs
        assert type(mu) is ndarray, mu
        assert type(Xps) is ndarray, Xps
        assert type(mu_criterion) is float, mu_criterion
        assert issubdtype(type(num_RC), integer), (type(num_RC), num_RC)
        assert num_RC >= min(num_RCs), (num_RC, num_RCs)
        assert num_RC <= max(num_RCs), (num_RC, num_RCs)
        self._series.append(kwargs)
        x: list = list(num_RCs)
        y: list = list(mu)
        dpg.bind_item_theme(
            dpg.add_scatter_series(
                x=x,
                y=y,
                label="µ",
                user_data=(
                    num_RCs,
                    mu,
                    Xps,
                ),
                parent=self._y_axis_1,
            ),
            themes.mu_Xps.mu,
        )
        dpg.bind_item_theme(
            dpg.add_scatter_series(
                x=[num_RC],
                y=[y[x.index(num_RC)]],
                parent=self._y_axis_1,
            ),
            themes.mu_Xps.mu_highlight,
        )
        dpg.bind_item_theme(
            dpg.add_line_series(
                x=x,
                y=y,
                user_data=(
                    num_RCs,
                    mu,
                    Xps,
                ),
                parent=self._y_axis_1,
            ),
            themes.mu_Xps.mu,
        )
        dpg.bind_item_theme(
            dpg.add_line_series(
                x=[x[0] - 2, x[-1] + 2],
                y=[mu_criterion, mu_criterion],
                label="µ-crit.",
                parent=self._y_axis_1,
            ),
            themes.mu_Xps.mu_criterion,
        )
        y = list(Xps)
        dpg.bind_item_theme(
            dpg.add_scatter_series(
                x=x,
                y=y,
                label="X²",
                parent=self._y_axis_2,
            ),
            themes.mu_Xps.Xps,
        )
        dpg.bind_item_theme(
            dpg.add_scatter_series(
                x=[num_RC],
                y=[y[x.index(num_RC)]],
                parent=self._y_axis_2,
            ),
            themes.mu_Xps.Xps_highlight,
        )
        dpg.bind_item_theme(
            dpg.add_line_series(
                x=x,
                y=y,
                parent=self._y_axis_2,
            ),
            themes.mu_Xps.Xps,
        )

    def adjust_limits(self):
        if not self.is_visible():
            self.queue_limits_adjustment()
            return
        elif self.are_limits_adjusted():
            return
        else:
            self.limits_adjusted()
        max_Xps: Optional[float] = None
        min_Xps: Optional[float] = None
        for kwargs in self._series:
            num_RCs: ndarray = kwargs["num_RCs"]
            Xps: ndarray = kwargs["Xps"]
            if max_Xps is None or max(Xps) > max_Xps:
                max_Xps = ceil(max(Xps))
            if min_Xps is None or min(Xps) > min_Xps:
                min_Xps = floor(min(Xps))
        if max_Xps is None:
            max_Xps = 1.0
        if min_Xps is None:
            min_Xps = 0.0
        dpg.split_frame()
        dpg.set_axis_limits(self._x_axis, ymin=min(num_RCs) - 1, ymax=max(num_RCs) + 1)
        dpg.set_axis_limits(self._y_axis_1, ymin=-0.1, ymax=1.1)
        dpg.set_axis_limits(self._y_axis_2, ymin=min_Xps - 0.1, ymax=max_Xps + 0.1)

    def copy_limits(self, other: Plot):
        src: int
        dst: int
        for src, dst in zip(
            [
                other._x_axis,
                other._y_axis_1,
                other._y_axis_2,
            ],
            [
                self._x_axis,
                self._y_axis_1,
                self._y_axis_2,
            ],
        ):
            limits: List[float] = dpg.get_axis_limits(src)
            dpg.set_axis_limits(dst, *limits)
        dpg.split_frame()
        dpg.set_axis_limits_auto(self._x_axis)
        dpg.set_axis_limits_auto(self._y_axis_1)
        dpg.set_axis_limits_auto(self._y_axis_2)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from typing import (
    Callable,
    List,
)
import dearpygui.dearpygui as dpg
from deareis.utility import calculate_window_position_dimensions
from deareis.tooltips import attach_tooltip
import deareis.tooltips as tooltips
from deareis.data.plotting import PlotExportSettings
from deareis.data import (
    DRTSettings,
    FitSettings,
    SimulationSettings,
    TestSettings,
)
from deareis.config import (
    DEFAULT_TEST_SETTINGS,
    DEFAULT_DRT_SETTINGS,
    DEFAULT_FIT_SETTINGS,
    DEFAULT_SIMULATION_SETTINGS,
    DEFAULT_PLOT_EXPORT_SETTINGS,
)
from deareis.gui.kramers_kronig import SettingsMenu as TestSettingsMenu
from deareis.gui.drt import SettingsMenu as DRTSettingsMenu
from deareis.gui.fitting import SettingsMenu as FitSettingsMenu
from deareis.gui.simulation import SettingsMenu as SimulationSettingsMenu
from deareis.gui.plotting.export import SettingsMenu as PlotExportSettingsMenu
from deareis.signals import Signal
import deareis.signals as signals


def section_spacer():
    dpg.add_spacer(height=8)


def general_settings(label_pad: int, state):
    with dpg.collapsing_header(label="General", default_open=True):
        auto_backup_interval: int = dpg.generate_uuid()

        def update_auto_backup_interval(value: int):
            state.config.auto_backup_interval = value

        with dpg.group(horizontal=True):
            dpg.add_text("Auto-backup interval".rjust(label_pad))
            attach_tooltip(tooltips.general.auto_backup_interval)
            dpg.add_input_int(
                default_value=state.config.auto_backup_interval,
                label="actions",
                min_value=0,
                min_clamped=True,
                step=0,
                on_enter=True,
                callback=lambda s, a, u: update_auto_backup_interval(a),
                width=-54,
                tag=auto_backup_interval,
            )
        section_spacer()


def kramers_kronig_tab_settings(label_pad: int, state) -> Callable:
    with dpg.collapsing_header(label="Kramers-Kronig tab", default_open=True):
        settings_menu: TestSettingsMenu = TestSettingsMenu(
            state.config.default_test_settings,
            label_pad,
            limited=True,
        )
        with dpg.group(horizontal=True):
            dpg.add_text("".rjust(label_pad))
            dpg.add_button(
                label="Restore defaults",
                callback=lambda s, a, u: settings_menu.set_settings(
                    DEFAULT_TEST_SETTINGS,
                ),
            )
        section_spacer()

    def callback():
        settings: TestSettings = settings_menu.get_settings()
        state.config.default_test_settings = settings
        signals.emit(Signal.APPLY_TEST_SETTINGS, settings=settings)

    return callback


def drt_tab_settings(label_pad: int, state) -> Callable:
    with dpg.collapsing_header(label="DRT analysis tab", default_open=True):
        settings_menu: DRTSettingsMenu = DRTSettingsMenu(
            state.config.default_drt_settings,
            label_pad,
        )
        with dpg.group(horizontal=True):
            dpg.add_text("".rjust(label_pad))
            dpg.add_button(
                label="Restore defaults",
                callback=lambda s, a, u: settings_menu.set_settings(
                    DEFAULT_DRT_SETTINGS
                ),
            )
        section_spacer()

    def callback():
        settings: DRTSettings = settings_menu.get_settings()
        state.config.default_drt_settings = settings
        signals.emit(Signal.APPLY_DRT_SETTINGS, settings=settings)

    return callback


def fitting_tab_settings(label_pad: int, state) -> Callable:
    with dpg.collapsing_header(label="Fitting tab", default_open=True):
        settings_menu: FitSettingsMenu = FitSettingsMenu(
            state.config.default_fit_settings,
            label_pad,
        )
        with dpg.group(horizontal=True):
            dpg.add_text("".rjust(label_pad))
            dpg.add_button(
                label="Restore defaults",
                callback=lambda s, a, u: settings_menu.set_settings(
                    DEFAULT_FIT_SETTINGS,
                ),
            )
        section_spacer()

    def callback():
        settings: FitSettings = settings_menu.get_settings()
        state.config.default_fit_settings = settings
        signals.emit(Signal.APPLY_FIT_SETTINGS, settings=settings)

    return callback


def simulation_tab_settings(label_pad: int, state) -> Callable:
    with dpg.collapsing_header(label="Simulation tab", default_open=True):
        settings_menu: SimulationSettingsMenu = SimulationSettingsMenu(
            state.config.default_simulation_settings,
            label_pad,
        )
        with dpg.group(horizontal=True):
            dpg.add_text("".rjust(label_pad))
            dpg.add_button(
                label="Restore defaults",
                callback=lambda s, a, u: settings_menu.set_settings(
                    DEFAULT_SIMULATION_SETTINGS,
                ),
            )
        section_spacer()

    def callback():
        settings: SimulationSettings = settings_menu.get_settings()
        state.config.default_simulation_settings = settings
        signals.emit(Signal.APPLY_SIMULATION_SETTINGS, settings=settings)

    return callback


def plotting_tab_settings(label_pad: int, state) -> Callable:
    with dpg.collapsing_header(label="Plotting tab - Export plot", default_open=True):
        settings_menu: PlotExportSettingsMenu = PlotExportSettingsMenu(
            state.config.default_plot_export_settings,
            label_pad,
            refresh_callback=None,
        )
        with dpg.group(horizontal=True):
            dpg.add_text("".rjust(label_pad))
            dpg.add_button(
                label="Restore defaults",
                callback=lambda s, a, u: settings_menu.set_settings(
                    DEFAULT_PLOT_EXPORT_SETTINGS,
                ),
            )
        section_spacer()

    def callback():
        settings: PlotExportSettings = settings_menu.get_settings()
        state.config.default_plot_export_settings = settings
        state.plot_exporter.set_settings(settings)

    return callback


def show_defaults_settings_window(state):
    x: int
    y: int
    w: int
    h: int
    x, y, w, h = calculate_window_position_dimensions(390, 540)

    window: int = dpg.generate_uuid()
    key_handler: int = dpg.generate_uuid()
    settings_update_callbacks: List[Callable] = []

    def close_window():
        for callback in settings_update_callbacks:
            callback()
        if dpg.does_item_exist(window):
            dpg.delete_item(window)
        if dpg.does_item_exist(key_handler):
            dpg.delete_item(key_handler)
        signals.emit(Signal.UNBLOCK_KEYBINDINGS)

    with dpg.handler_registry(tag=key_handler):
        dpg.add_key_release_handler(
            key=dpg.mvKey_Escape,
            callback=close_window,
        )

    with dpg.window(
        label="Settings - defaults",
        modal=True,
        pos=(
            x,
            y,
        ),
        width=w,
        height=h,
        no_resize=True,
        on_close=close_window,
        tag=window,
    ):
        label_pad: int = 24
        general_settings(label_pad, state)
        settings_update_callbacks.append(kramers_kronig_tab_settings(label_pad, state))
        settings_update_callbacks.append(drt_tab_settings(label_pad, state))
        settings_update_callbacks.append(fitting_tab_settings(label_pad, state))
        settings_update_callbacks.append(simulation_tab_settings(label_pad, state))
        settings_update_callbacks.append(plotting_tab_settings(label_pad, state))
    signals.emit(Signal.BLOCK_KEYBINDINGS, window=window, window_object=None)



# DearEIS is licensed under the GPLv3 or later (https://www.gnu.org/licenses/gpl-3.0.html).
# Copyright 2022 DearEIS developers
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# The licenses of DearEIS' dependencies and/or sources of portions of code are included in
# the LICENSES folder.

from multiprocessing import cpu_count
from typing import (
    Optional,
)
from pyimpspec.analysis.drt.bht import _get_default_num_procs
import deareis.api.drt as api
from deareis.data import (
    DRTResult,
    DRTSettings,
    DataSet,
    PlotSettings,
    Project,
)
from deareis.gui import ProjectTab
from deareis.signals import Signal
import deareis.signals as signals
from deareis.state import STATE


def select_drt_result(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    drt: Optional[DRTResult] = kwargs.get("drt")
    data: Optional[DataSet] = kwargs.get("data")
    if data is None or drt is None:
        return
    is_busy_message_visible: bool = STATE.is_busy_message_visible()
    if not is_busy_message_visible:
        signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Loading analysis result")
    project_tab.select_drt_result(drt, data)
    if not is_busy_message_visible:
        signals.emit(Signal.HIDE_BUSY_MESSAGE)


def delete_drt_result(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    drt: Optional[DRTResult] = kwargs.get("drt")
    data: Optional[DataSet] = kwargs.get("data")
    if data is None or drt is None:
        return
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Deleting analysis result")
    settings: Optional[PlotSettings] = project_tab.get_active_plot()
    update_plot: bool = (
        drt.uuid in settings.series_order if settings is not None else False
    )
    project.delete_drt(data, drt)
    project_tab.populate_drts(project, data)
    if settings is not None:
        project_tab.plotting_tab.populate_drts(
            project.get_all_drts(),
            project.get_data_sets(),
            settings,
        )
    if update_plot:
        signals.emit(Signal.SELECT_PLOT_SETTINGS, settings=settings)
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)


def apply_drt_settings(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    settings: Optional[DRTSettings] = kwargs.get("settings")
    if settings is None:
        return
    project_tab.set_drt_settings(settings)


def perform_drt(*args, **kwargs):
    project: Optional[Project] = STATE.get_active_project()
    project_tab: Optional[ProjectTab] = STATE.get_active_project_tab()
    if project is None or project_tab is None:
        return
    data: Optional[DataSet] = kwargs.get("data")
    settings: Optional[DRTSettings] = kwargs.get("settings")
    if data is None or settings is None:
        return
    assert (
        data.get_num_points() > 0
    ), "There are no data points to use to calculate the distribution of relaxation times!"
    num_procs: int = _get_default_num_procs()
    if num_procs > 1 and num_procs == cpu_count():
        num_procs -= 1
    signals.emit(Signal.SHOW_BUSY_MESSAGE, message="Performing analysis")
    drt: DRTResult = api.calculate_drt(
        data=data,
        settings=settings,
        num_procs=num_procs,
    )
    project.add_drt(data=data, drt=drt)
    project_tab.populate_drts(project, data)
    project_tab.plotting_tab.populate_drts(
        project.get_all_drts(),
        project.get_data_sets(),
        project_tab.get_active_plot(),
    )
    signals.emit(Signal.CREATE_PROJECT_SNAPSHOT)
    signals.emit(Signal.HIDE_BUSY_MESSAGE)
